-------------------------------------------------------------------------------
-- Localization
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("SimpleSelfRebuff")
local BS = AceLibrary("Babble-Spell-2.2")

L:RegisterTranslations("enUS", function() return {} end)

-------------------------------------------------------------------------------
-- Aliases & constants
-------------------------------------------------------------------------------

local tinsert = _G.tinsert
local db

local TEH_BUTTON_NAME = "SimpleSelfRebuffButton"

local trackings = {}
do
	local function initTrackings(...)
		for i=1,select('#', ...) do
			local t = select(i, ...)			
			trackings[BS:GetSpellIcon(t)] = BS[t]
		end
	end
	initTrackings(
		'Track Beasts',     'Track Demons', 'Track Dragonkin',
		'Track Elementals', 'Track Giants', 'Track Hidden',
		'Track Humanoids',  'Track Undead', 'Find Herbs',
		'Find Minerals',    'Find Treasure'
	)
end

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local registeredBuffs = {}
local categories = {}
local availableBuffs = {}
local actualBuffNames = {}
local actualBuffExpirations = {}

local buffToCast
local castButton
local spellScanned

-- local englishSpells
local active

-------------------------------------------------------------------------------
-- Addon declaration
-------------------------------------------------------------------------------

SimpleSelfRebuff = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0")
local SimpleSelfRebuff = SimpleSelfRebuff

do
	local m = SimpleSelfRebuff
	m.registeredBuffs = registeredBuffs
	m.categories = categories
	m.availableBuffs = availableBuffs
	m.actualBuffNames = actualBuffNames
	m.actualBuffExpirations = actualBuffExpirations	
	m.englishSpells = englishSpells
end

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function error(msg, ...)
	if select('#', ...) > 1 then
		msg = msg:format(...)
	end
	return _G.error(msg)
end

local function erase(...)
	for i=1,select('#', ...) do
		local t = select(i, ...)
		if type(t) == 'table' then
			for k in pairs(t) do
				t[k] = nil 
			end
		end
	end
end

-------------------------------------------------------------------------------
-- Initializing
-------------------------------------------------------------------------------

-- Addon setup
function SimpleSelfRebuff:OnInitialize()
	self.revision = tonumber(string.sub("$Revision: 41921 $", 12, -3))
	self.version = "r" .. SimpleSelfRebuff.revision
	self.date = string.sub("$Date: 2007-06-27 14:28:50 -0400 (Wed, 27 Jun 2007) $", 8, 17)

	self:RegisterDB("SimpleSelfRebuffDB")
	self:RegisterDefaults('profile', {
		categories = {
		},
	})

	db = self.db.profile

	self.options = {	
		type = 'group',
		args = {
			buffHeader = {
				type = 'header',
				name = ' ',
				order = 199,
			}
		}
	}
	
	self:RegisterChatCommand({"/ssr", "/simpleselfrebuff"}, self.options)
	
	if AceLibrary:HasInstance('AceDebug-2.0') then
		AceLibrary('AceDebug-2.0'):embed(SimpleSelfRebuff)
		SimpleSelfRebuff:SetDebugging(UnitName('player') == 'Adirelle')
	else
		function SimpleSelfRebuff.Debug() end
	end
		
end

function SimpleSelfRebuff:OnEnable(first)
	if first then
		self:ProcessRegistry()
	end

	self:RegisterBucketEvent('SPELLS_CHANGED', 1, 'ScanSpellbook')
	self:RegisterBucketEvent(
		{'PLAYER_DEAD', 'PLAYER_UNGHOST', 'PLAYER_ALIVE', 'PLAYER_UPDATE_RESTING', 'PLAYER_REGEN_DISABLED', 'PLAYER_REGEN_ENABLED'},
		0.2, 'CheckActivation'
	)
	
	self:ScanSpellbook()
end

-------------------------------------------------------------------------------
-- Buff registry
-------------------------------------------------------------------------------

do
	local buffFuncs = {}

	function SimpleSelfRebuff:RegisterBuffSetup(func)
		if type(func) == "function" then
			tinsert(buffFuncs, func)
		else
			error("Arg #1 to RegisterBuffs must be a function")
		end
	end
	
	local function registerCallback(...)
		return SimpleSelfRebuff:RegisterBuff(...)
	end

	function SimpleSelfRebuff:ProcessRegistry()
		if not buffFuncs then
			return
		end
		
		local selfClassName = select(2, UnitClass('player'))
		for i,func in ipairs(buffFuncs) do
			func(registerCallback, BS, selfClassName)
		end
		buffFuncs = nil

		-- Build categories
		for buffName, buffInfo in pairs(registeredBuffs) do
			if buffInfo.recast then
				local category = buffInfo.category
				if not categories[category] then
					categories[category] = {}
				end
				if not availableBuffs[category] then
					availableBuffs[category] = {}
				end
				tinsert(categories[category], buffName)
			end
		end
		
		-- Build categories options
		local opts = self.options.args
		for category, buffList in pairs(categories) do
			local category, buffList = category, buffList
			if #buffList == 1 then
				opts[category] = {
					type = 'toggle',
					name = category,
					desc = category,
					get = function()
						return db.categories[category] == category
					end,
					set = function(enable)
						db.categories[category] = enable and category or nil
						self:CheckRebuff()
					end,
					hidden = function()
						return not next(availableBuffs[category])
					end,
					order = 200,
				}
			else
				opts[category] = {
					type = 'text',
					name = category,
					desc = category,
					get = function()
						return db.categories[category]
					end,
					set = function(buffName)
						db.categories[category] = buffName
						self:CheckRebuff()
					end,
					validate = availableBuffs[category],
					hidden = function()
						return not next(availableBuffs[category])
					end,
					order = 200,
				}
			end
		end
	end
	
end

function SimpleSelfRebuff:RegisterBuff(name, category, recast)
	if registeredBuffs[name] then
		error("%q already registered", name)
	end
	registeredBuffs[name] = {
		name = name,
		category = category or name,
		recast = recast == nil or recast and true,
	}
end

-------------------------------------------------------------------------------
-- Activation
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckActivation()
	local shouldActivate = spellScanned and not (UnitIsDeadOrGhost('player') or InCombatLockdown() or IsResting()
	  or IsMounted() or IsFlying() or IsStealthed())

	if shouldActivate and not active then
		active = true
		self:Debug('Activated')
		
		self:RegisterBucketEvent('PLAYER_AURAS_CHANGED', 0.2, 'ScanPlayerAuras')
		self:ScanPlayerAuras()
		self:ScheduleRepeatingEvent('SimpleSelfRebuff_CheckRebuff', self.CheckRebuff, 1.5, self)
		self:CheckRebuff()

	elseif not shouldActivate and active then
		active = nil
		self:Debug('Disactivated')
		
		erase(actualBuffNames, actualBuffExpirations)

		self:CleanupPendingCast()		
		self:UnregisterBucketEvent('PLAYER_AURAS_CHANGED')
		self:CancelScheduledEvent('SimpleSelfRebuff_CheckRebuff')
		
	end
end

-------------------------------------------------------------------------------
-- Core
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckRebuff()

	local newBuffToCast = nil	
	local shortestBuffName, shortestBuffDuration
	
	for category in pairs(categories) do
		local expectedBuff = db.categories[category] 
		
		-- Always check cooldown
		local inCooldown
		if expectedBuff then
			local start, duration = GetSpellCooldown(expectedBuff)
			if start and start + duration > GetTime() then
				inCooldown = true
			end
		end
		
		-- Is we can cast the spell, check it further
		if expectedBuff and not inCooldown and IsUsableSpell(expectedBuff) then			
			local actualBuff = actualBuffNames[category] 
			
			-- Check if it is missing
			if not actualBuff then
				newBuffToCast = expectedBuff
				self:Debug('missing: %q', expectedBuff)
				break
				
			-- Check remaning time
			elseif actualBuff == expectedBuff and actualBuffExpirations[category] then
				local timeLeft = actualBuffExpirations[category] - GetTime()
				if timeLeft > 0 then
					if not shortestBuffDuration or timeLeft < shortestBuffDuration then
						shortestBuffDuration, shortestBuffName = timeLeft, actualBuff
					end
				end
			end
			
		end
	end
	
	if not newBuffToCast and shortestBuffName and shortestBuffDuration < 30 then
		newBuffToCast = shortestBuffName
		self:Debug('Buff to refresh: %q (%q)', newBuffToCast, shortestBuffDuration)
	end

	self:CleanupPendingCast()

	if newBuffToCast then
		if not castButton then
			castButton = CreateFrame("Button", TEH_BUTTON_NAME, UIParent, "SecureActionButtonTemplate")
			castButton:RegisterForClicks("LeftButtonUp")
			castButton:SetPoint("CENTER",0,0)
			castButton:Hide()

			castButton:SetAttribute('type', 'spell')
			castButton:SetAttribute('target', 'player')
		end
		
		buffToCast = newBuffToCast
		self:Debug('Binding mousewheel to %q', buffToCast)
		castButton:SetAttribute('spell', buffToCast)
		
		SetOverrideBindingClick(castButton, true, "MOUSEWHEELUP", TEH_BUTTON_NAME, "LeftButton")
		SetOverrideBindingClick(castButton, true, "MOUSEWHEELDOWN", TEH_BUTTON_NAME, "LeftButton")	
	end
	
end

function SimpleSelfRebuff:CleanupPendingCast()
	if buffToCast then
		self:Debug('Unbinding mousewheel', buffToCast)
		buffToCast = nil
		if castButton then
			ClearOverrideBindings(castButton)
		end
	end
end

-------------------------------------------------------------------------------
-- Spellbook handling
-------------------------------------------------------------------------------

--[[
englishSpells = setmetatable({}, {
	__index = function(self, key) 
		if BS:HasReverseTranslation(key) then
			rawset(self, key, BS:GetReverseTranslation(key))
		else
			rawset(self, key, key)
		end
		return self[key]
	end,
	__newindex = function() end,
})
--]]

function SimpleSelfRebuff:ScanSpellbook() 
	for k,v in pairs(availableBuffs) do
		erase(v)
	end
	local i = 1
	while true do
		local buffName = GetSpellName(i, BOOKTYPE_SPELL)
		if not buffName then
			break
		end
		local buffInfo = registeredBuffs[buffName]
		if buffInfo then
			availableBuffs[buffInfo.category][buffName] = buffName
		end
		spellScanned = true
		i = i + 1
	end
	self:CheckActivation()
end

-------------------------------------------------------------------------------
-- Player buff status tracking
-------------------------------------------------------------------------------

do
	local seen = {}
	
	function SimpleSelfRebuff:ScanPlayerAuras()
		erase(seen)
		local dirty
		
		for i = 1, 32 do
			local index = GetPlayerBuff(i, "HELPFUL")
			buffName = index and GetPlayerBuffName(index)		
			if buffName and registeredBuffs[buffName] then
				local category = registeredBuffs[buffName].category
				seen[category] = true
				if not actualBuffNames[category] then
					actualBuffNames[category] = buffName
					dirty = true
				end
				actualBuffExpirations[category] = GetTime() + (GetPlayerBuffTimeLeft(index) or 10000)
			end
		end
		
		local trackingTexture = GetTrackingTexture()
		local trackingName = trackingTexture and trackings[trackingTexture]
		if trackingName and registeredBuffs[trackingName] then
			local category = registeredBuffs[trackingName].category
			seen[category] = true
			actualBuffNames[category] = trackingName
			actualBuffExpirations[category] = GetTime() + 10000
			dirty = true
		end
		
		for category, actualBuff in pairs(actualBuffNames) do
			if not seen[category] then
				actualBuffNames[category] = nil
				actualBuffExpirations[category] = nil
				dirty = true
			end
		end
		
		if dirty then
			self:CheckRebuff()
		end
		
	end
	
end
